from __future__ import annotations

import contextlib
import functools
import inspect
import json
import logging
import os
import os.path
import warnings
from dataclasses import dataclass
from os import PathLike
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import pyarrow as pa

from chalk.features import Feature, FeatureConverter, Features, FeatureWrapper, unwrap_feature
from chalk.integrations.named import load_integration_variable
from chalk.sql._internal.chalk_query import ChalkQuery
from chalk.sql._internal.incremental import IncrementalSettings
from chalk.sql._internal.string_chalk_query import StringChalkQuery
from chalk.sql.finalized_query import FinalizedChalkQuery
from chalk.sql.protocols import BaseSQLSourceProtocol, ChalkQueryProtocol, StringChalkQueryProtocol, TableIngestProtocol
from chalk.utils.environment_parsing import env_var_bool
from chalk.utils.log_with_context import get_logger, get_logging_context
from chalk.utils.missing_dependency import missing_dependency_exception

if TYPE_CHECKING:
    from sqlalchemy.engine import URL, Connection, Engine
    from sqlalchemy.orm import Session
    from sqlalchemy.sql import Select
    from sqlalchemy.sql.elements import Label
    from sqlalchemy.types import TypeEngine

TTableIngestMixIn = TypeVar("TTableIngestMixIn", bound="TableIngestMixIn")
CHALK_QUERY_LOGGING = env_var_bool("CHALK_QUERY_LOGGING")

_logger = get_logger(__name__)


@dataclass
class TableIngestionPreferences:
    features: Type[Features]
    ignore_columns: Set[str]
    ignore_features: Set[str]
    require_columns: Set[str]
    require_features: Set[str]
    column_to_feature: Dict[str, str]
    cdc: Optional[Union[bool, IncrementalSettings]]


def _force_set_str(x: Optional[List[Any]]) -> Set[str]:
    return set() if x is None else set(map(str, x))


class UnsupportedEfficientExecutionError(ValueError):
    def __init__(self, msg: str, log_level: int) -> None:
        super().__init__(msg)
        self.log_level = log_level


def validate_dtypes_for_efficient_execution(stmt: Select, supported_types: Sequence[Type[TypeEngine]]):
    unsupported_columns = [
        x
        for x in stmt.column_descriptions
        # Using NOT IN rather than not isinstance() as subclasses may override the result processor,
        # which makes it not eligible for efficient execution
        if x["type"].__class__ not in supported_types
    ]
    if len(unsupported_columns) > 0:
        unsupported_columns_and_dtypes = [(x["name"], x["type"].__class__.__name__) for x in unsupported_columns]
        supported_dtypes = [x.__name__ for x in supported_types]
        formatted_supported = ", ".join(supported_dtypes)
        formatted_unsupported = ", ".join([f"{name}({dtype})" for (name, dtype) in unsupported_columns_and_dtypes])
        raise UnsupportedEfficientExecutionError(
            (
                "The SQL statement will be executed into SQLAlchemy objects, as the SQL query returns columns "
                "that cannot be loaded directly into a PyArrow table. For better performance, use only "
                f"the following types in your SQLAlchemy model: {formatted_supported}."
                f"The columns that contain unsupported types are: {formatted_unsupported}."
            ),
            log_level=logging.INFO,
        )


class TableIngestMixIn(TableIngestProtocol):
    ingested_tables: Dict[str, TableIngestionPreferences]

    def with_table(
        self: TTableIngestMixIn,
        *,
        name: str,
        features: Type[Union[Features, Any]],
        ignore_columns: Optional[List[str]] = None,
        ignore_features: Optional[List[Union[str, Any]]] = None,
        require_columns: Optional[List[str]] = None,
        require_features: Optional[List[Union[str, Any]]] = None,
        column_to_feature: Optional[Dict[str, Any]] = None,
        cdc: Optional[Union[bool, IncrementalSettings]] = None,
    ) -> TTableIngestMixIn:
        if name in self.ingested_tables:
            raise ValueError(f"The table {name} is ingested twice.")
        self.ingested_tables[name] = TableIngestionPreferences(
            features=features,
            ignore_columns=_force_set_str(ignore_columns),
            ignore_features=_force_set_str(ignore_features),
            require_columns=_force_set_str(require_columns),
            require_features=_force_set_str(require_features),
            column_to_feature={k: str(v) for k, v in (column_to_feature or {}).items()},
            cdc=cdc,
        )
        return self


class BaseSQLSource(BaseSQLSourceProtocol):
    registry: ClassVar[List["BaseSQLSource"]] = []

    def __init__(self, name: Optional[str], engine_args: Optional[Dict[str, Any]]):
        try:
            import sqlalchemy
        except ImportError:
            raise missing_dependency_exception("chalkpy[sql]")
        del sqlalchemy  # unused

        self._incremental_settings = None
        self._resolver_and_sqlfile_to_sqlstring: Dict[Tuple[str, str], str] = {}
        self.registry.append(self)
        self.name = name
        if engine_args is None:
            engine_args = {}
        if self.name is not None:
            for k, v in self._load_env_engine_args(name=self.name).items():
                engine_args.setdefault(k, v)
        engine_args.setdefault("pool_pre_ping", env_var_bool("USE_CLIENT_POOL_PRE_PING"))
        engine_args.setdefault("url", self.local_engine_url())
        engine_args.setdefault("echo", CHALK_QUERY_LOGGING)
        self.engine_args = engine_args

    def query_sql_file(
        self,
        path: Union[str, bytes, PathLike],
        fields: Optional[Mapping[str, Union[Feature, str, Any]]] = None,
        args: Optional[Mapping[str, object]] = None,
    ) -> StringChalkQueryProtocol:
        sql_string = None
        fqn = None

        try:
            fqn = get_logging_context().get("labels", {}).get("resolver_fqn")
            if fqn is not None:
                sql_string = self._resolver_and_sqlfile_to_sqlstring.get((fqn, str(path)))
        except Exception:
            pass

        uncached = sql_string is None

        if uncached:
            if os.path.isfile(path):
                with open(path) as f:
                    sql_string = f.read()
            else:
                caller_filename = inspect.stack()[1].filename
                dir_path = os.path.dirname(os.path.realpath(caller_filename))
                if isinstance(path, bytes):
                    path = path.decode("utf-8")
                relative_path = os.path.join(dir_path, path)
                if os.path.isfile(relative_path):
                    with open(relative_path) as f:
                        sql_string = f.read()
        if sql_string is None:
            raise FileNotFoundError(f"No such file: '{str(path)}'")

        if uncached and fqn is not None:
            self._resolver_and_sqlfile_to_sqlstring[(fqn, str(path))] = sql_string

        return self.query_string(
            query=sql_string,
            fields=fields,
            args=args,
        )

    def query_string(
        self,
        query: str,
        fields: Optional[Mapping[str, Union[Feature, Any]]] = None,
        args: Optional[Mapping[str, object]] = None,
    ) -> StringChalkQueryProtocol:
        fields = fields or {}
        fields = {f: unwrap_feature(v) if isinstance(v, FeatureWrapper) else v for (f, v) in fields.items()}
        return StringChalkQuery(source=self, query=query, fields=fields, params=args or {})

    def warm_up(self) -> None:
        from sqlalchemy import text

        with self.get_engine().connect() as cnx:
            cnx.execute(text("select 1"))

    def query(self, *entities: Any) -> ChalkQueryProtocol:
        targets: List[Label] = []
        features: Dict[str, Feature] = {}

        for e in entities:
            if isinstance(e, Features):
                _extract_features(e, e.__chalk_namespace__, targets, features)
            else:
                targets.append(e)

        return ChalkQuery(
            features=features,
            targets=targets,
            source=self,
        )

    def local_engine_url(self) -> URL:
        raise NotImplementedError

    def execute_query(
        self,
        finalized_query: FinalizedChalkQuery,
        columns_to_converters: Callable[[List[str]], Dict[str, FeatureConverter]],
        connection: Optional[Connection] = None,
    ) -> pa.Table:
        try:
            return self.execute_query_efficient(finalized_query, columns_to_converters, connection)
        except NotImplementedError:
            _logger.info(
                (
                    "The SQL statement will be executed into SQLAlchemy objects, as the database backend does "
                    "not support a more efficient execution mechanism."
                )
            )
            pass
        except UnsupportedEfficientExecutionError as e:
            log_level = e.log_level
            _logger.log(log_level, str(e))

        return self.execute_query_inefficient(finalized_query, columns_to_converters, connection)

    def execute_query_inefficient(
        self,
        finalized_query: FinalizedChalkQuery,
        columns_to_converters: Callable[[List[str]], Dict[str, FeatureConverter]],
        connection: Optional[Connection],
    ):
        with self.get_engine().connect() if connection is None else contextlib.nullcontext(connection) as cnx:
            res = cnx.execute(finalized_query.query, finalized_query.params)
            desc = res.cursor.description
            result_columns: list[str] = [col[0] for col in desc]
            converters = columns_to_converters(result_columns)
            data: Dict[str, List[Any]] = {}
            for v in converters.keys():
                # Create an entry for the columns, so the pyarrow table will have the correct columns even
                # if there is no data
                data[v] = []
            for row in res.all():
                for k, v in zip(result_columns, row):
                    if k not in data:
                        # We are not interested in this column
                        continue
                    data[k].append(converters[k].from_rich_to_primitive(v, missing_value_strategy="default_or_allow"))
            # Only keep the columns provided by the schema
            schema = pa.schema([pa.field(k, v.pyarrow_dtype) for (k, v) in converters.items()])
            return pa.Table.from_pydict(data, schema=schema)

    def execute_query_efficient(
        self,
        finalized_query: FinalizedChalkQuery,
        columns_to_converters: Callable[[List[str]], Dict[str, FeatureConverter]],
        connection: Optional[Connection],
    ) -> pa.Table:
        raise NotImplementedError()

    def compile_query(
        self,
        finalized_query: FinalizedChalkQuery,
        paramstyle: Optional[str] = None,
    ) -> Tuple[str, Sequence[Any], Dict[str, Any]]:
        """Compile a query into a string and the bindparams"""
        dialect = self.local_engine_url().get_dialect()(paramstyle=paramstyle)
        query = finalized_query.query.params(finalized_query.params)
        compiled_query = query.compile(dialect=dialect)
        query_string = compiled_query.string
        return query_string, compiled_query.positiontup or [], compiled_query.params

    def _load_env_engine_args(self, name: str) -> Mapping[str, Any]:
        """
        Loads additional engine arguments from env var "{name}_ENGINE_ARGUMENTS"
        """

        extra_args = load_integration_variable(integration_name=name, name="ENGINE_ARGUMENTS")
        if extra_args is None:
            return {}
        else:
            extra_args = json.loads(extra_args)
            assert isinstance(extra_args, dict), "ENGINE_ARGUMENTS must be a JSON object"
            return extra_args

    def get_engine(self, additional_engine_args: Optional[Dict[str, Any]] = None) -> Engine:
        if additional_engine_args is not None:
            # Deprecating the additional engine args, and instead requiring they are specified when creating the engine,
            # since otherwise the engine won't be cached
            # If we need different types of engines for a given SQL source, we should probably add explicit methods
            # e.g. get_read_engine, get_write_engine, etc...
            warnings.warn(
                DeprecationWarning(
                    (
                        "The argument `additional_engine_args` is deprecated for the method `SQLSource.get_engine`. "
                        "Instead, specify the arguments when constructing the SQL source -- "
                        "e.g. `PostgresSQLSource(engine_args=...)`"
                    )
                )
            )
        isolation_level = self.engine_args.get("isolation_level")
        if isolation_level == "AUTOCOMMIT":
            warnings.warn(
                UserWarning(
                    (
                        f"The SQL engine '{self.name}' is being created with the AUTOCOMMIT transaction isolation level, which helps improve "
                        "performance for SELECT statements by avoiding unnecessary transactions. If a different transaction level is needed for an "
                        "individual connection, use the `execution_options` method when retrieving a connection -- e.g. "
                        "`with get_engine().connect().execution_options(isolation_level='REPEATABLE READ') as cnx: ...`. "
                        "For more information, please see "
                        "https://docs.sqlalchemy.org/en/20/core/connections.html#setting-isolation-level-or-dbapi-autocommit-for-a-connection"
                    )
                )
            )

        if additional_engine_args is None:
            return self._get_engine_default()
        else:
            return self._get_engine_with_args(additional_engine_args)

    def _get_engine_with_args(self, additional_engine_args: Dict[str, Any]) -> Engine:
        from sqlalchemy.engine import create_engine

        args = dict(self.engine_args)
        args.update(additional_engine_args)

        return create_engine(**args)

    @functools.lru_cache(None)
    def _get_engine_default(self) -> Engine:
        from sqlalchemy.engine import create_engine

        return create_engine(**self.engine_args)

    def raw_session(self) -> Session:
        from sqlalchemy.orm import Session

        warnings.warn(
            DeprecationWarning(
                (
                    "The method `raw_session()` is deprecated. Instead, please construct a session directly "
                    "from the underlying engine -- for example:"
                    "`from sqlalchemy.orm import Session; with Session(sql_source.get_engine()) as session: ...`"
                )
            )
        )
        return Session(self.get_engine())


def _extract_features(feature_set: Features, prefix: str, targets: List[Label], features: Dict[str, Feature]):
    import sqlalchemy.sql.functions
    import sqlalchemy.sql.schema
    from sqlalchemy import column
    from sqlalchemy.orm import InstrumentedAttribute

    for f in feature_set.features:
        assert f.attribute_name is not None
        try:
            feature_value = getattr(feature_set, f.attribute_name)
        except AttributeError:
            continue
        if f.is_has_many:
            raise ValueError(
                f"Feature '{prefix}.{f.name}' is a nested has-many feature, which is not supported when querying from SQL"
            )
        root_fqn = f"{prefix}.{f.name}"
        if isinstance(feature_value, Features):
            # Nested sub-features
            _extract_features(feature_value, root_fqn, targets, features)
        elif isinstance(feature_value, str):
            # Treat it as a column name
            features[root_fqn] = Feature.from_root_fqn(root_fqn)
            targets.append(column(feature_value).label(root_fqn))
        elif isinstance(feature_value, (sqlalchemy.sql.functions.GenericFunction, InstrumentedAttribute)):
            features[root_fqn] = Feature.from_root_fqn(root_fqn)
            targets.append(feature_value.label(root_fqn))
        elif isinstance(feature_value, sqlalchemy.sql.schema.Column):
            features[root_fqn] = Feature.from_root_fqn(root_fqn)
            targets.append(feature_value.label(root_fqn))
        else:
            raise TypeError(
                (
                    f"Feature '{root_fqn}' has an unsupported value of type '{type(feature_value)}' for SQL queries. "
                    "All values must be a column, a column name string, a nested Features class, or an SQLAlchemy function"
                )
            )



from __future__ import annotations

import dataclasses
import decimal
import enum
from datetime import date, datetime, time, timedelta
from typing import TYPE_CHECKING, Dict, FrozenSet, List, Set, Tuple, Type, TypedDict, Union, cast

import attrs
import pyarrow as pa
from pydantic import BaseModel
from typing_extensions import Annotated, get_args, get_origin, get_type_hints, is_typeddict

from chalk.features._encoding.primitive import TPrimitive
from chalk.utils.collections import is_namedtuple, unwrap_optional_and_annotated_if_needed
from chalk.utils.enum import get_enum_value_type
from chalk.utils.missing_dependency import missing_dependency_exception

if TYPE_CHECKING:
    import polars as pl


__all__ = ["pyarrow_to_primitive", "pyarrow_to_polars", "rich_to_pyarrow"]


def rich_to_pyarrow(python_type: Type, name: str) -> pa.DataType:
    # Polars seems to allow optional for any dtype, so we ignore it when computing dtypes
    python_type = unwrap_optional_and_annotated_if_needed(python_type)
    origin = get_origin(python_type)
    if origin is not None:
        # Handling namedtuples above as structs before tuples, to ensure a namedtuple is not treated like
        # a list
        args = get_args(python_type)
        if origin == Annotated:
            if len(args) < 1:
                raise TypeError(
                    "Annotated types must contain the underlying type as the first argument -- e.g. Annotated[int, 'annotation']"
                )
            return rich_to_pyarrow(args[0], name)
        if origin in (list, List, set, Set, frozenset, FrozenSet):
            typ_name = origin.__name__
            if len(args) == 0:
                raise TypeError(
                    f"{typ_name} features must be annotated with the type of the element -- e.g. {typ_name}[int]"
                )
            if len(args) > 1:
                raise TypeError(
                    f"{typ_name} annotations should only take one argument -- e.g. {typ_name}[int]. Instead, got {typ_name}[{', '.join(args)}]"
                )
            arg = args[0]
            return pa.list_(rich_to_pyarrow(arg, name=f"{name}[]"))
        if origin in (tuple, Tuple):
            if len(args) == 0:
                raise TypeError(
                    "Tuple features must be annotated with the type of the tuple element -- e.g. `Tuple[int, ...]`"
                )
            if len(args) == 2 and args[1] is ...:
                # Treat a variable-sized, homogenous tuple like a list
                arg = args[0]
                return pa.list_(rich_to_pyarrow(arg, name=f"{name}[]"))
            raise TypeError(
                (
                    "Tuple features must have a fixed type and be variable-length tuples (e.g. `Tuple[int, ...]`). "
                    " If you would like a fixed-length of potentially different types, used a NamedTuple."
                )
            )
        raise TypeError(f"Unsupported varardic type annotation: {origin}")
    if not isinstance(python_type, type):
        raise TypeError(f"Type annotations must be a type. Instead, got {python_type}.")
    if issubclass(python_type, enum.Enum):
        # For enums, require all members to have the same type
        return rich_to_pyarrow(get_enum_value_type(python_type), name)
        # First, handle the recursive types -- list, tuple, typeddict, namedtuple, dataclass, pydantic model
    if (
        dataclasses.is_dataclass(python_type)
        or is_namedtuple(python_type)
        or is_typeddict(python_type)
        or attrs.has(python_type)
        or issubclass(python_type, BaseModel)
    ):
        annotations = get_type_hints(python_type)
        fields: List[pa.Field] = []
        for field_name, type_annotation in annotations.items():
            underlying_dtype = rich_to_pyarrow(type_annotation, name=f"{name}.{field_name}")
            fields.append(pa.field(field_name, underlying_dtype))
        return pa.struct(fields)
    if issubclass(python_type, str):
        return pa.string()
    if issubclass(python_type, bool):
        return pa.bool_()
    if issubclass(python_type, int):
        return pa.int64()
    if issubclass(python_type, float):
        return pa.float64()
    if issubclass(python_type, datetime):
        return pa.timestamp("us", "UTC")
    if issubclass(python_type, date):
        return pa.date64()
    if issubclass(python_type, time):
        return pa.time64("us")
    if issubclass(python_type, timedelta):
        return pa.duration("us")
    if issubclass(python_type, bytes):
        return pa.binary()
    if issubclass(python_type, decimal.Decimal):
        # Using a string for decimals, since polars
        # does not support decimal types
        return pa.utf8()

    raise TypeError(
        (
            f"Unable to determine the PyArrow type for field '{name}' with type `{python_type}`. "
            "Please set the `dtype` attribute when defining the feature"
        )
    )


def pyarrow_to_primitive(pyarrow_typ: pa.DataType, name: str) -> Type[TPrimitive]:
    if pa.types.is_null(pyarrow_typ):
        return cast(Type[TPrimitive], None)
    if pa.types.is_boolean(pyarrow_typ):
        return cast(Type[TPrimitive], bool)
    if pa.types.is_unsigned_integer(pyarrow_typ) or pa.types.is_signed_integer(pyarrow_typ):
        return cast(Type[TPrimitive], int)
    if pa.types.is_floating(pyarrow_typ):
        return cast(Type[TPrimitive], float)
    if pa.types.is_time(pyarrow_typ):
        return cast(Type[TPrimitive], time)
    if pa.types.is_timestamp(pyarrow_typ):
        return cast(Type[TPrimitive], datetime)
    if pa.types.is_date(pyarrow_typ):
        return cast(Type[TPrimitive], date)
    if pa.types.is_duration(pyarrow_typ):
        return cast(Type[TPrimitive], timedelta)
    if (
        pa.types.is_binary(pyarrow_typ)
        or pa.types.is_fixed_size_binary(pyarrow_typ)
        or pa.types.is_large_binary(pyarrow_typ)
    ):
        return cast(Type[TPrimitive], bytes)
    if pa.types.is_string(pyarrow_typ) or pa.types.is_large_string(pyarrow_typ):
        return cast(Type[TPrimitive], str)
    if pa.types.is_list(pyarrow_typ) or pa.types.is_large_list(pyarrow_typ) or pa.types.is_fixed_size_list(pyarrow_typ):
        pyarrow_typ = cast(Union[pa.ListType, pa.LargeListType, pa.FixedSizeListType], pyarrow_typ)
        underlying = pyarrow_typ.value_type
        return cast(Type[TPrimitive], List[pyarrow_to_primitive(underlying, name=f"{name}[]")])
    if pa.types.is_struct(pyarrow_typ):
        pyarrow_typ = cast(pa.StructType, pyarrow_typ)
        # Dynamically making a TypedDict so members will be parsed with the correct types by cattrs
        annotations: Dict[str, Type[TPrimitive]] = {}
        schema = pa.schema(pyarrow_typ)
        for sub_name, sub_typ in zip(schema.names, schema.types):
            annotations[sub_name] = pyarrow_to_primitive(sub_typ, name=f"{name}.{sub_name}")
        typed_dict_cls = TypedDict(f"__chalk_struct__{name}", annotations)  # type: ignore
        return cast(Type[TPrimitive], typed_dict_cls)
    raise TypeError(f"Unsupported PyArrow type {pyarrow_typ} for field {name}")


def pyarrow_to_polars(pa_type: pa.DataType, name: str) -> pl.PolarsDataType:
    """Convert a PyArrow data type into a Polars DataType

    Args:
        pa_type: The PyArrow data type
        name: A name, which is printed in error messages
    """
    try:
        import polars as pl
    except ImportError:
        raise missing_dependency_exception("chalkpy[runtime]")
    if pa.types.is_null(pa_type):
        return pl.Null()
    if pa.types.is_boolean(pa_type):
        return pl.Boolean()
    if pa.types.is_int8(pa_type):
        return pl.Int8()
    if pa.types.is_int16(pa_type):
        return pl.Int16()
    if pa.types.is_int32(pa_type):
        return pl.Int32()
    if pa.types.is_int64(pa_type):
        return pl.Int64()
    if pa.types.is_uint8(pa_type):
        return pl.UInt8()
    if pa.types.is_uint16(pa_type):
        return pl.UInt16()
    if pa.types.is_uint32(pa_type):
        return pl.UInt32()
    if pa.types.is_uint64(pa_type):
        return pl.UInt64()
    if pa.types.is_float16(pa_type):
        return pl.Float32()
    if pa.types.is_float32(pa_type):
        return pl.Float32()
    if pa.types.is_float64(pa_type):
        return pl.Float64()
    if pa.types.is_time(pa_type):
        return pl.Time()
    if pa.types.is_timestamp(pa_type):
        pa_type = cast(pa.TimestampType, pa_type)
        assert pa_type.unit in ("ms", "us", "ns")
        return pl.Datetime(pa_type.unit, pa_type.tz)
    if pa.types.is_date(pa_type):
        return pl.Date()
    if pa.types.is_duration(pa_type):
        pa_type = cast(pa.DurationType, pa_type)
        assert pa_type.unit in ("ms", "us", "ns")
        return pl.Duration(pa_type.unit)
    if pa.types.is_binary(pa_type) or pa.types.is_fixed_size_binary(pa_type) or pa.types.is_large_binary(pa_type):
        return pl.Binary()
    if pa.types.is_string(pa_type) or pa.types.is_large_string(pa_type):
        return pl.Utf8()
    if pa.types.is_list(pa_type) or pa.types.is_large_list(pa_type) or pa.types.is_fixed_size_list(pa_type):
        pa_type = cast(Union[pa.ListType, pa.LargeListType, pa.FixedSizeListType], pa_type)
        underlying = pa_type.value_type
        return pl.List(pyarrow_to_polars(underlying, name=f"{name}[]"))
    if pa.types.is_struct(pa_type):
        pa_type = cast(pa.StructType, pa_type)
        schema = pa.schema(pa_type)

        fields = [
            pl.Field(sub_name, pyarrow_to_polars(sub_typ, name=f"{name}.{sub_name}"))
            for sub_name, sub_typ in zip(schema.names, schema.types)
        ]
        return pl.Struct(fields)
    raise TypeError(f"Unsupported PyArrow type {pa_type} for field {name}")



from typing import Any

from typing_extensions import TypeAlias

MachineType: TypeAlias = str
"""The type of machine to use.

You can optionally specify that resolvers need to run
on a machine other than the default. Must be configured
in your deployment.
"""

AnyDataclass: TypeAlias = Any
"""Any class decorated by `@dataclass`.

There isn't a base class for `dataclass`, so we use this
`TypeAlias` to refer to indicate any class decorated with
`@dataclass`.
"""



import argparse
import json
import logging
import os
import sys
import traceback
from pathlib import Path
from typing import List, Optional

from pydantic import ValidationError

from chalk.config.project_config import load_project_config
from chalk.importer import import_all_files
from chalk.parsed.user_types_to_json import get_registered_types_as_json
from chalk.utils.stubgen import configure_stubgen_argparse, run_stubgen


def _get_list_results(directory: Optional[str]):
    scope_to = Path(directory or os.getcwd())
    try:
        failed = import_all_files()
        return get_registered_types_as_json(scope_to, failed)
    except Exception:
        ex_type, ex_value, ex_traceback = sys.exc_info()
        assert ex_type is not None
        relevant_traceback = f"""{os.linesep.join(traceback.format_tb(ex_traceback))}
\n{ex_type.__name__}: {str(ex_value)}
"""
        return json.dumps(dict(failed=[dict(traceback=relevant_traceback)]), indent=2)


def dump_cmd(filename: str, directory: Optional[str]):
    with open(filename, "w") as f:
        f.write(_get_list_results(directory))


def list_cmd(directory: str):
    print(_get_list_results(directory))


def config_cmd():
    json_response: str
    try:
        model = load_project_config()
    except ValidationError as e:
        json_response = json.dumps({"error": str(e)})
    else:
        if model is None:
            print("No `chalk.yaml` configuration file found")
            return
        json_response = model.json()

    print(json_response)


def cli(args_override: Optional[List[str]] = None):
    parser = argparse.ArgumentParser(
        prog="Chalk Python CLI",
        description=(
            "You typically do not need to invoke this utility directly. "
            "Prefer https://github.com/chalk-ai/cli instead."
        ),
    )
    parser.add_argument("--log-level", help="Print debug info", nargs="?")
    subparsers = parser.add_subparsers(dest="command")
    subparsers.add_parser("config", help="Print the config for the current project")
    list_parser = subparsers.add_parser("list", help="List the resolvers for the current project and print as json")
    list_parser.add_argument("--directory", help="Scope to this directory", nargs="?")
    dump_parser = subparsers.add_parser("dump", help="Write the resolvers for the current project to the given file")
    dump_parser.add_argument("filename", help="Write to this file")
    dump_parser.add_argument("--directory", help="Scope to this directory", nargs="?")
    stubgen_parser = subparsers.add_parser("stubgen", help="Generate type stubs for feature set classes")
    configure_stubgen_argparse(stubgen_parser)
    args = parser.parse_args(args_override)
    if args.log_level:
        level = getattr(logging, args.log_level.upper())
        logging.basicConfig(
            format="%(asctime)s %(levelname)s: %(message)s",
            datefmt="%m/%d/%Y %I:%M:%S %p",
            level=level,
        )

    if args.command == "list":
        list_cmd(args.directory)

    elif args.command == "config":
        config_cmd()

    elif args.command == "stubgen":
        run_stubgen(args)

    elif args.command == "dump":
        dump_cmd(args.filename, args.directory)
    else:
        parser.print_help(sys.stderr)
        return 1
    return 0


if __name__ == "__main__":
    sys.exit(cli())



from __future__ import annotations

import ast
import copy
import functools
import inspect
import re
import sys
import textwrap
import threading
import types
from datetime import datetime, timedelta
from itertools import islice
from typing import Any, Callable, Dict, List, Mapping, Optional, Sequence, Tuple, Type, TypeVar, Union, cast, overload

from chalk.features._class_property import classproperty, classproperty_support
from chalk.features.feature_field import Feature, _VersionInfo
from chalk.features.feature_set import Features, FeatureSetBase
from chalk.features.feature_time import feature_time
from chalk.features.feature_wrapper import FeatureWrapper, unwrap_feature
from chalk.features.tag import Tags
from chalk.serialization.parsed_annotation import ParsedAnnotation
from chalk.streams import Windowed
from chalk.utils.collections import ensure_tuple
from chalk.utils.duration import Duration, parse_chalk_duration
from chalk.utils.metaprogramming import MISSING, create_fn, field_assign, set_new_attribute
from chalk.utils.string import removeprefix, to_snake_case

T = TypeVar("T")

__all__ = ["features"]

GENERATED_OBSERVED_AT_NAME = "__chalk_observed_at__"


@overload
def features(
    *,
    owner: Optional[str] = None,
    tags: Optional[Tags] = None,
    etl_offline_to_online: bool = False,
    max_staleness: Optional[Duration] = None,
    name: Optional[str] = None,
) -> Callable[[Type[T]], Type[T]]:
    ...


@overload
def features(cls: Type[T]) -> Type[T]:
    ...


def features(
    cls: Optional[Type[T]] = None,
    *,
    owner: Optional[str] = None,
    tags: Optional[Tags] = None,
    etl_offline_to_online: bool = False,
    max_staleness: Optional[Duration] = None,
    name: Optional[str] = None,
) -> Union[Callable[[Type[T]], Type[T]], Type[T]]:
    """Chalk lets you spell out your features directly in Python.

    Features are namespaced to a `FeatureSet`.
    To create a new `FeatureSet`, apply the `@features`
    decorator to a Python class with typed attributes.
    A `FeatureSet` is constructed and functions much like
    Python's own `dataclass`.

    Parameters
    ----------
    owner
        The individual or team responsible for these features.
        The Chalk Dashboard will display this field, and alerts
        can be routed to owners.
    tags
        Added metadata for features for use in filtering, aggregations,
        and visualizations. For example, you can use tags to assign
        features to a team and find all features for a given team.
    etl_offline_to_online
        When `True`, Chalk copies this feature into the online environment
        when it is computed in offline resolvers.
        Setting `etl_offline_to_online` on a feature class assigns it to all features on the
        class which do not explicitly specify `etl_offline_to_online`.
    max_staleness
        When a feature is expensive or slow to compute, you may wish to cache its value.
        Chalk uses the terminology "maximum staleness" to describe how recently a feature
        value needs to have been computed to be returned without re-running a resolver.
        Assigning a `max_staleness` to the feature class assigns it to all features on the
        class which do not explicitly specify a `max_staleness` value of their own.

    Other Parameters
    ----------------
    cls
        The decorated class. You shouldn't need to pass this argument.
    name
        The name for the feature set. By default, the name of a feature is
        taken from the name of the attribute on the class, prefixed with
        the camel-cased name of the class.

    Examples
    --------
    >>> @features(
    ...     owner="andy@chalk.ai",
    ...     max_staleness="30m",
    ...     etl_offline_to_online=True,
    ...     tags="user-group",
    ... )
    ... class User:
    ...     id: str
    ...     name: str | None
    ...     location: LatLng
    """

    def wrap(c: Type[T]) -> Type[T]:
        nonlocal max_staleness
        namespace = name or to_snake_case(c.__name__)
        if name is not None and re.sub(r"[^a-z_0-9]", "", namespace) != namespace:
            raise ValueError(
                (
                    f"Namespace must be composed of lower-case alpha-numeric characters and '_'. Provided namespace "
                    f"'{namespace}' for class '{c.__name__}' contains invalid characters."
                )
            )
        if name is not None and len(namespace) == 0:
            raise ValueError(f"Namespace cannot be an empty string, but is for the class '{c.__name__}'.")
        if max_staleness is None:
            max_staleness = timedelta(0)
        if isinstance(max_staleness, str):
            max_staleness = parse_chalk_duration(max_staleness)
        updated_class = _process_class(
            cls=c,
            owner=owner,
            tags=ensure_tuple(tags),
            etl_offline_to_online=etl_offline_to_online,
            max_staleness=max_staleness,
            namespace=namespace,
        )
        assert issubclass(updated_class, Features)
        FeatureSetBase.registry[updated_class.__chalk_namespace__] = updated_class
        if FeatureSetBase.hook is not None:
            FeatureSetBase.hook(cast(Features, updated_class))

        return cast(Type[T], updated_class)

    # See if we're being called as @features or @features().
    if cls is None:
        # We're called with parens.
        return wrap

    # We're called as @features without parens.
    return wrap(cls)


def _discover_feature(
    features: Sequence[Feature], name: str, *conditions: Callable[[Feature], bool]
) -> Optional[Feature]:
    for cond in conditions:
        filtered_features = [c for c in features if cond(c)]
        if len(filtered_features) == 1:
            return filtered_features[0]
        if len(filtered_features) > 1:
            assert filtered_features[0].features_cls is not None
            feature_cls_name = filtered_features[0].features_cls.__name__
            raise ValueError(
                f"Multiple {name} features are not supported in {feature_cls_name}: "
                + ", ".join(f"{feature_cls_name}.{x.name}" for x in filtered_features)
            )
    return None


def _init_param(name: str, annotation: Any, _locals: Dict[str, Any]):
    # Return the __init__ parameter string for this field.  For
    # example, the equivalent of 'x:int=3' (except instead of 'int',
    # reference a variable set to int, and instead of '3', reference a
    # variable set to 3).
    annotation_name = f"_type_{name}"
    _locals[annotation_name] = annotation
    return f"{name}:{annotation_name}=MISSING"


def _setattr_fn(bidirectional_alias: Dict[str, str], _globals: Dict[str, Any]):
    return create_fn(
        name="__setattr__",
        args=["self", "key", "value"],
        body=[
            f"alias = {bidirectional_alias}",
            "if key in alias:",
            "    super(self.__class__, self).__setattr__(alias[key], value)",
            "return super(self.__class__, self).__setattr__(key, value)",
        ],
        _globals=_globals,
        _locals={},
        return_type=None,
    )


def _getattribute_fn(_globals: Dict[str, Any]):
    # If calling getattr() on an instance for a feature name,
    # do NOT return a class-level FeatureWrapper.
    # Instead, raise an attribute error

    return create_fn(
        name="__getattribute__",
        args=["self", "attribute_name"],
        body=[
            "o = object.__getattribute__(self, attribute_name)",
            "if isinstance(o, FeatureWrapper):",
            "    raise AttributeError(f'Feature \\'{attribute_name}\\' is not defined on this instance of class \\'{type(self).__name__}\\'')",
            "return o",
        ],
        _globals=_globals,
        _locals={
            "FeatureWrapper": FeatureWrapper,
        },
        return_type=Any,
    )


def _init_fn(
    feature_names_and_annotations: List[Tuple[str, Any]],
    alias_from_to: Mapping[str, str],
    _globals: Dict[str, Any],
):
    _locals = {
        "MISSING": MISSING,
    }
    _init_params = [_init_param(name, annotation, _locals) for (name, annotation) in feature_names_and_annotations]
    _init_params.append(_init_param(GENERATED_OBSERVED_AT_NAME, datetime, _locals))

    # Not using "self" for the self name in case if there is a feature called "self"
    self_name = "__chalk_self__"
    body_lines: List[str] = []
    for from_, to in alias_from_to.items():
        body_lines.extend(
            [
                f"if {from_} is not MISSING and {to} is not MISSING:",
                f"    raise ValueError('The features \\'{from_}\\' and \\'{to}\\' are aliases of each other. Only one can be specified, but both were given.')",
                f"{from_} = {to} if {from_} is MISSING else {from_}",
            ]
        )
    body_lines.extend([field_assign(name, name, self_name) for (name, _) in feature_names_and_annotations])

    # Check to see whether the GENERATED_OBSERVED_AT_NAME is the name of the timestamp feature
    # We must include GENERATED_OBSERVED_AT_NAME in the __init__ signature, even if there is a timestamp
    # feature defined and named, because we cannot parse the class annotations until after the module is fully
    # loaded (otherwise we'll get circular errors when trying to resolve forward reference annotations)
    # So, we ALWAYS include the GENERATED_OBSERVED_AT_NAME as part of the signature, and then error if
    # that is not the name of the timestamp feature
    error_msg = f"{self_name}.__class__.__name__ + \".__init__() got unexpected value for argument '{GENERATED_OBSERVED_AT_NAME}'\""
    has_explicit_ts_feature = f"{self_name}.__chalk_ts__.attribute_name != '{GENERATED_OBSERVED_AT_NAME}'"
    specified_val_for_generated_observed_at = f"{GENERATED_OBSERVED_AT_NAME} is not MISSING"
    # Raising a TypeError to be consistent with an unexpected argument message
    body_lines.append(
        f"if {specified_val_for_generated_observed_at} and {has_explicit_ts_feature}: raise TypeError({error_msg})"
    )
    body_lines.append(field_assign(GENERATED_OBSERVED_AT_NAME, GENERATED_OBSERVED_AT_NAME, self_name))

    return create_fn(
        name="__init__",
        args=[self_name] + _init_params,
        body=body_lines,
        _locals=_locals,
        _globals=_globals,
        return_type=None,
    )


def _parse_tags(ts: str) -> List[str]:
    ts = re.sub(",", " ", ts)
    ts = re.sub(" +", " ", ts.strip())
    return [xx.strip() for xx in ts.split(" ")]


def _get_field(
    cls: Type,
    annotation_name: str,
    comments: Mapping[str, str],
    class_owner: Optional[str],
    class_tags: Optional[Tuple[str, ...]],
    class_etl_offline_to_online: bool,
    class_max_staleness: timedelta,
    namespace: str,
) -> Feature:
    # Return a Field object for this field name and type.  ClassVars and
    # InitVars are also returned, but marked as such (see f._field_type).
    # default_kw_only is the value of kw_only to use if there isn't a field()
    # that defines it.

    # If the default value isn't derived from Field, then it's only a
    # normal default value.  Convert it to a Field().
    default = getattr(cls, annotation_name, ...)

    if isinstance(default, Feature):
        # The feature was set like x: int = Feature(...)
        f = default
        if f.version is not None:
            f.version.base_name = f._name or annotation_name
            f.name = f.version.name_for_version(f.version.default)
        if f.is_name_set():
            if "." in f.name:
                raise ValueError(
                    f"Custom feature names cannot contain a dot, but the feature '{f.name}' on the class '{cls.__name__}' includes a dot. You might consider using a has-one feature instead."
                )
            elif " " in f.name:
                raise ValueError(
                    f"Custom feature names cannot contain spaces, but the feature '{f.name}' on the class '{cls.__name__}' includes a space."
                )
    elif isinstance(default, Windowed):
        # The feature was set like x: Windowed[int] = windowed()
        # Convert it to a Feature
        f = default._to_feature(bucket=None)
    else:
        # The feature was not set explicitly
        if isinstance(default, types.MemberDescriptorType):
            # This is a field in __slots__, so it has no default value.
            default = ...
        f = Feature(name=annotation_name, namespace=namespace, default=default)

    # Only at this point do we know the name and the type.  Set them.
    f.namespace = namespace
    f.typ = ParsedAnnotation(cls, annotation_name)

    f.features_cls = cls
    f.attribute_name = annotation_name
    _process_field(f, comments, class_owner, class_tags, class_etl_offline_to_online, class_max_staleness)
    return f


def _process_field(
    f: Feature,
    comments: Mapping[str, str],
    class_owner: Optional[str],
    class_tags: Optional[Tuple[str, ...]],
    class_etl_offline_to_online: bool,
    class_max_staleness: timedelta,
) -> Feature:
    comment_for_feature = comments.get(f.attribute_name)
    comment_based_description = None
    comment_based_owner = None
    comment_based_tags = None

    if comment_for_feature is not None:
        comment_lines = []
        for line in comment_for_feature.splitlines():
            stripped_line = line.strip()
            if stripped_line.startswith(":owner:"):
                comment_based_owner = removeprefix(stripped_line, ":owner:").strip()
            elif stripped_line.startswith(":tags:"):
                parsed = _parse_tags(removeprefix(stripped_line, ":tags:"))
                if len(parsed) > 0:
                    if comment_based_tags is None:
                        comment_based_tags = parsed
                    else:
                        comment_based_tags.extend(parsed)
            else:
                comment_lines.append(line)

        comment_based_description = "\n".join(comment_lines)

    if f.description is None and comment_based_description is not None:
        f.description = comment_based_description

    if comment_based_tags is not None:
        if f.tags is None:
            f.tags = comment_based_tags
        else:
            f.tags.extend(comment_based_tags)

    if class_tags is not None:
        if f.tags is None:
            f.tags = list(class_tags)
        else:
            f.tags.extend(class_tags)

    if f.owner is not None and comment_based_owner is not None:
        raise ValueError(
            (
                f"Owner for {f.namespace}.{f.name} specified both on the feature and in the comment. "
                f"Please use only one of these two."
            )
        )
    elif f.owner is None:
        f.owner = comment_based_owner or class_owner

    # Using the private variable because the max_staleness is a read-only property
    if f._max_staleness is ...:
        f._max_staleness = class_max_staleness

    # Using the private variable because the etl_offline_to_online is a read-only property
    if f._etl_offline_to_online is None:
        f._etl_offline_to_online = False if class_etl_offline_to_online is None else class_etl_offline_to_online
    return f


def _recursive_repr(user_function: Callable[[T], str]) -> Callable[[T], str]:
    # Decorator to make a repr function return "..." for a recursive
    # call.
    repr_running = set()

    @functools.wraps(user_function)
    def wrapper(self: T):
        key = id(self), threading.get_ident()
        if key in repr_running:
            return "..."
        repr_running.add(key)
        try:
            result = user_function(self)
        finally:
            repr_running.discard(key)
        return result

    return wrapper


def _repr_fn(_globals: Dict[str, Any]):
    tuples = "((f.attribute_name, getattr(self, f.attribute_name, MISSING)) for f in self.features)"
    fn = create_fn(
        name="__repr__",
        args=["self"],
        body=[
            'return f"{self.__class__.__qualname__}(" + '
            + f"', '.join(f'{{x[0]}}={{x[1]}}' for x in {tuples} if x[1] != MISSING)"
            + '+ ")"'
        ],
        _globals=_globals,
        _locals={"MISSING": MISSING},
    )
    return _recursive_repr(fn)


def _eq_fn(_globals: Dict[str, Any]):
    cmp_str = "all(getattr(self, f.attribute_name, MISSING) == getattr(other, f.attribute_name, MISSING) for f in self.features)"
    return create_fn(
        name="__eq__",
        args=["self", "other"],
        body=[
            "if not isinstance(other, type(self)):",
            "    return NotImplemented",
            f"return {cmp_str}",
        ],
        _globals=_globals,
        _locals={
            "MISSING": MISSING,
        },
    )


def _len_fn(_globals: Dict[str, Any]):
    return create_fn(
        name="__len__",
        args=["self"],
        body=[
            "count = 0",
            f"for f in self.features:",
            "    if not f.no_display and hasattr(self, f.attribute_name):",
            "        count += 1",
            "return count",
        ],
        _globals=_globals,
        _locals={},
    )


def _iter_fn(_globals: Mapping[str, Any]):
    return create_fn(
        "__iter__",
        args=["self"],
        body=[
            f"for __chalk_f__ in self.features:",
            f"    if hasattr(self, __chalk_f__.attribute_name) and type(__chalk_f__).__name__ == 'Feature' and not __chalk_f__.is_has_one and not __chalk_f__.is_has_many and not __chalk_f__.no_display:",
            f"        yield __chalk_f__.fqn, getattr(self, __chalk_f__.attribute_name)",
        ],
    )


def _parse_annotation_comments(cls: Type) -> Mapping[str, str]:
    source = textwrap.dedent(inspect.getsource(cls))

    source_lines = source.splitlines()
    tree = ast.parse(source)
    if len(tree.body) != 1:
        return {}

    comments_for_annotations: Dict[str, str] = {}
    class_def = tree.body[0]
    if isinstance(class_def, ast.ClassDef):
        for stmt in class_def.body:
            if isinstance(stmt, ast.AnnAssign) and isinstance(stmt.target, ast.Name):
                line = stmt.lineno - 2
                comments: List[str] = []
                while line >= 0 and source_lines[line].strip().startswith("#"):
                    comment = source_lines[line].strip().strip("#").strip()
                    comments.insert(0, comment)
                    line -= 1

                if len(comments) > 0:
                    comments_for_annotations[stmt.target.id] = textwrap.dedent("\n".join(comments))

    return comments_for_annotations


def _process_class(
    cls: Type[T],
    owner: Optional[str],
    tags: Tuple[str, ...],
    etl_offline_to_online: bool,
    max_staleness: timedelta,
    namespace: str,
) -> Type[T]:
    raw_cls_annotations = cls.__dict__.get("__annotations__", {})

    cls_annotations: Dict[str, Any] = {}
    for name, annotation in raw_cls_annotations.items():
        if isinstance(annotation, Windowed):
            # NOTE: For Windowed resolvers, both the Annotation and the value are instances of Windowed,
            # unlike normal features whose annotation is the underlying type, and the value is an instance
            # of FeatureWrapper. So both `annotation` and `wind` should be instances of Windowed
            # In the future, we should use a subclass of Windowed, rather than an instance, for the type annotation,
            # similar to what we do for Features
            wind = getattr(cls, name, None)
            if wind is None or not isinstance(wind, Windowed):
                assert annotation._kind is not None
                raise TypeError(
                    (
                        f"Windowed feature '{namespace}.{name}' is missing windows. "
                        f"To create a windowed feature, use "
                        f"'{name}: Windowed[{annotation._kind.__name__}] = windowed('10m', '30m')."
                    )
                )
            wind.kind = annotation.kind
            wind._name = name
            annotation._buckets = wind._buckets
            for bucket in wind._buckets:
                # Make pseudofeatures for each bucket of the window
                feat = wind._to_feature(bucket=bucket)
                setattr(cls, feat.name, feat)
                # For the psuedofeatures, which track an individual bucket, the correct annotation is the
                # underlying annotation, not Windowed[underlying], since it's only one value
                cls_annotations[feat.name] = wind.kind

        cls_annotations[name] = annotation

    if cls.__module__ in sys.modules:
        _globals = sys.modules[cls.__module__].__dict__
    else:
        _globals = {}

    # Find feature times that weren't annotated.
    for name, member in inspect.getmembers(cls):
        if name not in cls_annotations and isinstance(member, Windowed):
            raise TypeError(f"Windowed feature '{namespace}.{name}' is missing an annotation, like 'Windowed[str]'")
        if name not in cls_annotations and isinstance(member, Feature):
            # All feature types need annotations, except for datetimes, which we can automatically infer
            if not member.is_typ_set() or not member.typ.is_parsed:
                raise TypeError(
                    f"Feature '{namespace}.{name}' is missing an annotation. Please add one, like '{name}: str = ...'"
                )
            cls_annotations[name] = member.typ.annotation

    cls.__annotations__ = cls_annotations

    # If we pass this function something that isn't a class, it could raise
    try:
        comments = _parse_annotation_comments(cls)
    except:
        comments = {}

    cls_fields: List[Feature] = []

    def __chalk_primary__(cls: Type[Features]):
        return _discover_feature(
            cls_fields,
            "primary",
            lambda f: f._primary is True,
            lambda f: f.typ.is_primary,
            lambda f: (
                f.name == "id" and not f.has_resolved_join and not f._is_feature_time and not f.typ.is_feature_time
            ),
        )

    set_new_attribute(cls, "__chalk_primary__", value=classproperty(__chalk_primary__, cached=True))

    def __chalk_ts__(cls: Type[Features]) -> Feature:
        # Not using `f.is_feature_time` as that would create an infinite recursion, since
        # `.is_feature_time` accesses `__chalk_ts__`
        ts_feature: Optional[Feature] = _discover_feature(
            cls_fields,
            "feature time",
            lambda f: f._is_feature_time is True,
            lambda f: f.typ.is_feature_time,
            lambda f: f.name == "ts" and not f.has_resolved_join and not f._primary and not f.typ.is_primary,
        )
        if ts_feature is None:
            return unwrap_feature(getattr(cls, GENERATED_OBSERVED_AT_NAME))
        return ts_feature

    set_new_attribute(cls=cls, name="__chalk_ts__", value=classproperty(__chalk_ts__, cached=True))

    def __chalk_observed_at__(cls: Type[Features]) -> FeatureWrapper:
        ts_feature: Optional[Feature] = _discover_feature(
            cls_fields,
            "feature time",
            # Not using `f.is_feature_time` as that would create an infinite recursion, since
            # `.is_feature_time` accesses `__chalk_ts__` which can access this function
            lambda f: f._is_feature_time is True,
            lambda f: f.typ.is_feature_time,
            lambda f: f.name == "ts" and not f.has_resolved_join and not f._primary and not f.typ.is_primary,
        )
        if ts_feature is not None:
            if ts_feature.attribute_name != GENERATED_OBSERVED_AT_NAME:
                raise AttributeError(f"Object {cls.__name__} has no attribute '{GENERATED_OBSERVED_AT_NAME}")
        # If the timestamp feature is still none, then synthesize one on first use
        ts_feature = feature_time()
        assert ts_feature is not None
        ts_feature.name = GENERATED_OBSERVED_AT_NAME
        ts_feature.attribute_name = GENERATED_OBSERVED_AT_NAME
        ts_feature.namespace = cls.__chalk_namespace__
        ts_feature.features_cls = cls
        ts_feature.is_autogenerated = True
        cls.__annotations__[GENERATED_OBSERVED_AT_NAME] = datetime
        _process_field(
            ts_feature,
            comments={},
            class_owner=cls.__chalk_owner__,
            class_tags=tuple(cls.__chalk_tags__),
            class_etl_offline_to_online=cls.__chalk_etl_offline_to_online__,
            class_max_staleness=cls.__chalk_max_staleness__,
        )

        return FeatureWrapper(ts_feature)

    set_new_attribute(
        cls=cls,
        name=GENERATED_OBSERVED_AT_NAME,
        value=classproperty(__chalk_observed_at__, cached=True, bind_to_instances=False),
    )

    def __features__(cls: Features) -> List[Feature]:
        features = list(cls_fields)
        if cls.__chalk_ts__ not in features:
            assert cls.__chalk_ts__ is not None
            features.append(cls.__chalk_ts__)
        return features

    set_new_attribute(cls=cls, name="features", value=classproperty(__features__, cached=True))
    set_new_attribute(cls=cls, name="__repr__", value=_repr_fn(_globals=_globals))
    set_new_attribute(cls=cls, name="__eq__", value=_eq_fn(_globals=_globals))
    set_new_attribute(cls=cls, name="__hash__", value=None)
    set_new_attribute(cls=cls, name="__iter__", value=_iter_fn(_globals=_globals))

    set_new_attribute(cls=cls, name="namespace", value=namespace)
    set_new_attribute(cls=cls, name="__chalk_namespace__", value=namespace)
    set_new_attribute(cls=cls, name="__chalk_owner__", value=owner)
    set_new_attribute(cls=cls, name="__chalk_tags__", value=list(tags))
    set_new_attribute(cls=cls, name="__chalk_max_staleness__", value=max_staleness)
    set_new_attribute(
        cls=cls,
        name="__chalk_etl_offline_to_online__",
        value=etl_offline_to_online,
    )
    set_new_attribute(cls=cls, name="__is_features__", value=True)
    set_new_attribute(cls=cls, name="__len__", value=_len_fn(_globals=_globals))
    set_new_attribute(cls=cls, name="__getattribute__", value=_getattribute_fn(_globals=_globals))
    # Moving this line lower causes all kinds of problems.
    cls = classproperty_support(cls)

    # Parse the fields after we have the correct `cls` set
    cls_fields.extend(
        _get_field(
            cls=cls,
            annotation_name=name,
            comments=comments,
            class_owner=owner,
            class_tags=tags,
            class_etl_offline_to_online=etl_offline_to_online,
            class_max_staleness=max_staleness,
            namespace=namespace,
        )
        for name in cls_annotations
    )

    alias_from_to: Dict[str, str] = {}

    # We're appending to this list
    for f in islice(cls_fields, len(cls_fields)):
        if f.version is None:
            continue
        alias_from_to[f.attribute_name] = f"{f.attribute_name}_v{f.version.default}"

        for i in range(1, f.version.maximum + 1):
            f_i = copy.copy(f)
            f_i.name = f.version.name_for_version(i)
            f_i.attribute_name = f"{f.attribute_name}_v{i}"
            f_i.version = _VersionInfo(
                version=i,
                maximum=f.version.maximum,
                default=f.version.default,
                reference=f.version.reference,
            )
            f.version.reference[i] = f_i
            cls_fields.append(f_i)

            # The default feature already exists.
            f_i.no_display = i == f.version.default

            if f_i.attribute_name in cls_annotations:
                assert f_i.features_cls is not None
                raise ValueError(
                    (
                        f"The class '{f_i.features_cls.__name__}' "
                        f"has an existing annotation '{f_i.attribute_name}' "
                        "that collides with a versioned feature. Please remove the existing "
                        "annotation, or lower the version."
                    )
                )

    set_new_attribute(
        cls=cls,
        name="__setattr__",
        value=_setattr_fn(
            bidirectional_alias=dict({v: k for k, v in alias_from_to.items()}, **alias_from_to),
            _globals=_globals,
        ),
    )
    set_new_attribute(
        cls=cls,
        name="__init__",
        value=_init_fn(
            feature_names_and_annotations=[(x.attribute_name, x.typ.annotation) for x in cls_fields],
            alias_from_to=alias_from_to,
            _globals=_globals,
        ),
    )

    for f in cls_fields:
        assert f.attribute_name is not None
        # Wrap all class features with FeatureWrapper
        setattr(cls, f.attribute_name, FeatureWrapper(f))

    return cls



from __future__ import annotations

import base64
import collections.abc
from datetime import date, datetime, time, timedelta
from typing import Any, List, Optional, Type, Union, cast

import cattrs
import dateutil.parser
import isodate
from typing_extensions import get_args, get_origin, get_type_hints, is_typeddict

from chalk.features._encoding.primitive import TPrimitive
from chalk.utils.json import TJSON

__all__ = ["unstructure_primitive_to_json", "structure_json_to_primitive"]


_json_converter = cattrs.Converter()


def unstructure_primitive_to_json(val: TPrimitive) -> TJSON:
    return _json_converter.unstructure(val)


def structure_json_to_primitive(val: Union[TJSON, TPrimitive], typ: Type[TPrimitive]) -> TPrimitive:
    return _json_converter.structure(val, typ)


#######
# Dicts
#######

# All structs get mapped to a typeddict primitive type
# However, on the wire, we serialize typedicts as lists by field order, similar
# to how structs are commonly serialized by values
# Hence, the JSON converter needs to convert all dicts into lists of the values
# In py3.7+, all dicts are ordered. So, simply iterating over the dict values is sufficient
_json_converter.register_unstructure_hook(dict, lambda x: [_json_converter.unstructure(y) for y in x.values()])


def _structure_typeddict(obj: Any, typ: Type):
    type_hints = get_type_hints(typ)
    if obj is None:
        return {
            field_name: _json_converter.structure(None, type_hint) for (field_name, type_hint) in type_hints.items()
        }
    if isinstance(obj, collections.abc.Mapping):
        # If given a dict, assume it is the primitive type being passed in as the json type
        return {
            field_name: _json_converter.structure(obj.get(field_name), type_hint)
            for (field_name, type_hint) in type_hints.items()
        }
    if not isinstance(obj, collections.abc.Sequence):
        raise TypeError(f"Expected structs to be serialized as lists. Object `{obj}` is not a sequence.")

    if len(type_hints) != len(obj):
        raise TypeError(
            f"Unable to structure object `{obj}` of size {len(obj)} into type `{typ.__name__}` of size {len(type_hints)}. Size mismatch."
        )

    kwargs = {
        field_name: None if x is None else _json_converter.structure(x, type_hints[field_name])
        for (x, field_name) in zip(obj, type_hints.keys())
    }
    return kwargs


_json_converter.register_structure_hook_func(is_typeddict, _structure_typeddict)


#######
# Lists
#######

# The default unstructure hook is fine
# However, when structuring, we want to allow None for annotations of list[...]


def _structure_list(obj: Optional[List], typ: Type[List]) -> List:
    args = get_args(typ)

    if len(args) < 1:
        raise TypeError(
            f"{typ} types must be parameterized with the type of the contained value -- for example, `{typ}[int]`"
        )
    if len(args) > 1:
        raise TypeError(f"{typ} should be parameterized with only one type")
    if obj is None:
        return cast(List, None)
    if not isinstance(obj, (list, tuple)):
        raise TypeError(f"Expected a list, Object `{obj}` is not a list.")

    inner_typ = args[0]
    return [_json_converter.structure(x, inner_typ) for x in obj]


def _is_list(typ: Type):
    origin = get_origin(typ)
    return origin in (list, List)


_json_converter.register_structure_hook_func(_is_list, _structure_list)


######
# Date
######
def _is_date(x: Type):
    return isinstance(x, type) and issubclass(x, date) and not issubclass(x, datetime)


_json_converter.register_unstructure_hook_func(
    _is_date,
    lambda x: x.isoformat(),
)


def _structure_date(obj: Any, typ: Type):
    if isinstance(obj, datetime):
        return obj.date()
    if isinstance(obj, date):
        return obj
    if not isinstance(obj, str):
        raise TypeError(
            f"Date values must be serialized as ISO strings. Instead, received value '{obj}' of type `{type(obj).__name__}`"
        )
    return isodate.parse_date(obj)


_json_converter.register_structure_hook_func(
    _is_date,
    _structure_date,
)


##########
# Datetime
##########


_json_converter.register_unstructure_hook(datetime, lambda x: x.isoformat())


def _structure_datetime(obj: Any, typ: Type):
    if obj is None:
        return None
    if isinstance(obj, datetime):
        return obj
    if isinstance(obj, date):
        # Upgrade to a datetime
        return datetime.combine(obj, time())
    if not isinstance(obj, str):
        raise TypeError(
            f"Datetime values must be serialized as ISO strings. Instead, received value '{obj}' of type `{type(obj).__name__}`"
        )
    return dateutil.parser.parse(obj)


_json_converter.register_structure_hook(datetime, _structure_datetime)

######
# Time
######

_json_converter.register_unstructure_hook(time, lambda x: x.isoformat())


def _structure_time(obj: Any, typ: Type):
    if obj is None:
        return None
    if isinstance(obj, time):
        return obj
    if not isinstance(obj, str):
        raise TypeError(
            f"Time values must be serialized as ISO strings. Instead, received value '{obj}' of type `{type(obj).__name__}`"
        )
    return isodate.parse_time(obj)


_json_converter.register_structure_hook(time, _structure_time)

########
# Binary
########

_json_converter.register_unstructure_hook(bytes, lambda x: base64.b64encode(x).decode("utf8"))


def _structure_bytes(obj: Any, typ: Type):
    if obj is None:
        return None
    if isinstance(obj, str):
        return base64.b64decode(obj)
    if isinstance(obj, bytes):
        return obj
    raise TypeError(
        f"Byte values must be bytes objects or Base64-encoded strings. Instead, received value '{obj}' of type `{type(obj).__name__}`"
    )


_json_converter.register_structure_hook(bytes, _structure_bytes)

_json_converter.register_unstructure_hook(timedelta, isodate.duration_isoformat)

###########
# Timedelta
###########


def _structure_timedelta(obj: Any, typ: Type):
    if obj is None:
        return None
    if isinstance(obj, timedelta):
        return obj
    if not isinstance(obj, str):
        raise TypeError(
            f"Timedelta values should be serialized as strings. Instead, received value '{obj}' of type `{type(obj).__name__}`"
        )
    return isodate.parse_duration(obj)


_json_converter.register_structure_hook(timedelta, _structure_timedelta)

####################
# Int/float/str/bool
####################


def _structure_basic(obj: Any, typ: Type):
    if obj is None:
        # Always allow None, even if the field is non-optional
        return None
    if issubclass(typ, bool):
        # For booleans, we are more strict than just doing a simple cast, since
        # bool("random string") or bool(100) should raise an exception, not be True
        if obj in (1, True):
            return True
        if obj in (0, False):
            return False
        raise TypeError(f"Cannot convert '{obj}' to a Boolean. Valid values are 1, True, 0, or False.")
    if issubclass(typ, (int, float)):
        if typ(obj) != obj:
            raise TypeError(f"Cannot cast '{obj}' to a {typ} without losing precision")
    return typ(obj)


_json_converter.register_structure_hook(int, _structure_basic)
_json_converter.register_structure_hook(float, _structure_basic)
_json_converter.register_structure_hook(str, _structure_basic)
_json_converter.register_structure_hook(bool, _structure_basic)



from collections.abc import Mapping as MappingABC
from typing import Any, Dict, List, Mapping, Optional, Tuple

from chalk.features import Feature, FeatureNotFoundException, Features, ensure_feature


def recursive_encode(inputs: Mapping) -> Tuple[Dict[str, Any], List]:
    all_warnings = []
    encoded_inputs = {}
    for fqn, v in inputs.items():
        try:
            feature = ensure_feature(fqn)
        except FeatureNotFoundException:
            all_warnings.append(f"Input {fqn} not recognized. JSON encoding {fqn} and requesting anyways")
            encoded_inputs[fqn] = v
            continue

        if feature.is_has_many:
            _validate_has_many_value(v, feature)

            has_many_result = []
            for item in v:
                result, inner_warnings = recursive_encode(item)
                all_warnings = all_warnings + inner_warnings
                has_many_result.append(result)

            encoded_inputs[feature.root_fqn] = has_many_result
        else:
            encoded_inputs[feature.root_fqn] = feature.converter.from_rich_to_json(
                v,
                missing_value_strategy="error",
            )

    return encoded_inputs, all_warnings


def _validate_has_many_value(v: Any, feature: Feature):
    if not isinstance(v, list):
        raise TypeError(f"has-many feature '{feature.fqn}' must be a list, but got {type(v).__name__}")

    for item in v:
        if not isinstance(item, MappingABC):
            raise TypeError(
                f"has-many feature '{feature.fqn}' must be a list of Mapping, but got a list of {type(item).__name__}"
            )



from __future__ import annotations

import os
from typing import TYPE_CHECKING, Any, Dict, Optional

from chalk.integrations.named import load_integration_variable
from chalk.sql._internal.sql_source import BaseSQLSource
from chalk.utils.missing_dependency import missing_dependency_exception

if TYPE_CHECKING:
    from sqlalchemy.engine.url import URL


class RedshiftSourceImpl(BaseSQLSource):
    def __init__(
        self,
        host: Optional[str] = None,
        db: Optional[str] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        name: Optional[str] = None,
        engine_args: Optional[Dict[str, Any]] = None,
    ):
        try:
            import redshift_connector
        except ImportError:
            raise missing_dependency_exception("chalkpy[redshift]")
        del redshift_connector
        self.host = host or load_integration_variable(name="REDSHIFT_HOST", integration_name=name)
        self.db = db or load_integration_variable(name="REDSHIFT_DB", integration_name=name)
        self.user = user or load_integration_variable(name="REDSHIFT_USER", integration_name=name)
        self.password = password or load_integration_variable(name="REDSHIFT_PASSWORD", integration_name=name)
        if engine_args is None:
            engine_args = {}
        engine_args.setdefault("pool_size", 20)
        engine_args.setdefault("max_overflow", 60)
        engine_args.setdefault(
            "connect_args",
            {
                "keepalives": 1,
                "keepalives_idle": 30,
                "keepalives_interval": 10,
                "keepalives_count": 5,
            },
        )
        # We set the default isolation level to autocommit since the SQL sources are read-only, and thus
        # transactions are not needed
        # Setting the isolation level on the engine, instead of the connection, avoids
        # a DBAPI statement to reset the transactional level back to the default before returning the
        # connection to the pool
        engine_args.setdefault("isolation_level", os.environ.get("CHALK_SQL_ISOLATION_LEVEL", "AUTOCOMMIT"))
        BaseSQLSource.__init__(self, name=name, engine_args=engine_args)

    def local_engine_url(self) -> URL:
        from sqlalchemy.engine.url import URL

        return URL.create(
            drivername="redshift+psycopg2",
            username=self.user,
            password=self.password,
            host=self.host,
            database=self.db,
        )



import dataclasses
from datetime import timedelta
from typing import Literal, Optional


@dataclasses.dataclass
class IncrementalSettings:
    """Incremental settings for Chalk SQL queries.

    In `"row"` mode:
    `incremental_column` MUST be set.

    Returns the results represented by this query as a list (like `.all()`), but modifies the query to
    only return "new" results, by adding a clause that looks like:

        ```
        "WHERE <incremental_column> >= <previous_latest_row_timestamp> - <lookback_period>"
        ```

    In `"group"` mode:
    `incremental_column` MUST be set.

    Returns the results represented by this query as a list (like `.all()`), but modifies the query to
    only results from "groups" which have changed since the last run of the query.

    This works by (1) parsing your query, (2) finding the "group keys", (3) selecting only changed groups.
    Concretely:

        ```
        SELECT user_id, sum(amount) as sum_amount
        FROM payments
        GROUP BY user_id
        ```

    would be rewritten like this:

        ```
        SELECT user_id, sum(amount) as sum_amount
        FROM payments
        WHERE user_id in (
            SELECT DISTINCT(user_id)
            FROM payments WHERE created_at >= <previous_latest_row_timestamp> - <lookback_period>
        )
        GROUP BY user_id
        ```

    In "parameter" mode:
    `incremental_column` WILL BE IGNORED.

    This mode is for cases where you want full control of incrementalization. Chalk will not manipulate your query.
    Chalk will include a query parameter named `"chalk_incremental_timestamp"`. Depending on your SQL
    dialect, you can use this value to incrementalize your query with `:chalk_incremental_timestamp` or
    `%(chalk_incremental_timestamp)s`.
    """

    mode: Literal["row", "group", "parameter"] = "row"  # "row" or "group" or "parameter"

    lookback_period: Optional[timedelta] = None
    """The amount of overlap to check for late-arriving rows."""

    incremental_column: Optional[str] = None
    """The column on which to incrementalize."""



from __future__ import annotations

import os
from typing import TYPE_CHECKING, Any, Dict, Optional

from chalk.sql._internal.sql_source import BaseSQLSource
from chalk.utils.missing_dependency import missing_dependency_exception

if TYPE_CHECKING:
    from sqlalchemy.engine.url import URL


class CloudSQLSourceImpl(BaseSQLSource):
    def __init__(
        self,
        *,
        instance_name: Optional[str] = None,
        db: Optional[str] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        name: Optional[str] = None,
        engine_args: Optional[Dict[str, Any]] = None,
    ):
        try:
            import psycopg2
        except ImportError:
            raise missing_dependency_exception("chalkpy[postgresql]")
        del psycopg2  # unused
        prefix = name + "_" if name else ""
        self.instance_name = instance_name or os.getenv(prefix + "CLOUDSQL_INSTANCE_NAME")
        self.db = db or os.getenv(prefix + "CLOUDSQL_DATABASE")
        self.user = user or os.getenv(prefix + "CLOUDSQL_USER")
        self.password = password or os.getenv(prefix + "CLOUDSQL_PASSWORD")
        if engine_args is None:
            engine_args = {}
        engine_args.setdefault("pool_size", 20)
        engine_args.setdefault("max_overflow", 60)
        engine_args.setdefault(
            "connect_args",
            {
                "keepalives": 1,
                "keepalives_idle": 30,
                "keepalives_interval": 10,
                "keepalives_count": 5,
            },
        )
        # We set the default isolation level to autocommit since the SQL sources are read-only, and thus
        # transactions are not needed
        # Setting the isolation level on the engine, instead of the connection, avoids
        # a DBAPI statement to reset the transactional level back to the default before returning the
        # connection to the pool
        engine_args.setdefault("isolation_level", os.environ.get("CHALK_SQL_ISOLATION_LEVEL", "AUTOCOMMIT"))
        BaseSQLSource.__init__(self, name=name, engine_args=engine_args)

    def local_engine_url(self) -> URL:
        from sqlalchemy.engine.url import URL

        return URL.create(
            drivername="postgresql",
            username=self.user,
            password=self.password,
            host="",
            query={"host": "{}/{}/.s.PGSQL.5432".format("/cloudsql", self.instance_name)},
            database=self.db,
        )



from typing import Optional


def get_ipython_string() -> Optional[str]:
    """
    :return: "ZMQInteractiveShell" for jupyter notebook, "TerminalInteractiveShell" for ipython in terminal, None otherwise.
    """
    try:
        # I know this has a redline under it... we'll catch the NameError as a Falsy condition below
        # noinspection PyUnresolvedReferences
        shell = get_ipython().__class__.__name__
        return shell
    except NameError:
        return None  # Probably standard Python interpreter


def is_notebook() -> bool:
    """:return: true if run inside a Jupyter notebook"""
    shell = get_ipython_string()
    return shell == "ZMQInteractiveShell"



from __future__ import annotations

import os
from os import PathLike
from typing import TYPE_CHECKING, Any, Dict, Optional, TypeVar, Union

from chalk.sql._internal.sql_source import BaseSQLSource, TableIngestMixIn
from chalk.sql.protocols import SQLSourceWithTableIngestProtocol
from chalk.utils.missing_dependency import missing_dependency_exception

if TYPE_CHECKING:
    from sqlalchemy.engine.url import URL


T = TypeVar("T")


class SQLiteSourceImpl(TableIngestMixIn, BaseSQLSource, SQLSourceWithTableIngestProtocol):
    def __init__(
        self,
        name: Optional[str] = None,
        filename: Optional[Union[PathLike, str]] = None,
        engine_args: Optional[Dict[str, Any]] = None,
    ):
        self.ingested_tables: Dict[str, Any] = {}
        self.filename = filename
        if engine_args is None:
            engine_args = {}
        # We set the default isolation level to autocommit since the SQL sources are read-only, and thus
        # transactions are not needed
        # Setting the isolation level on the engine, instead of the connection, avoids
        # a DBAPI statement to reset the transactional level back to the default before returning the
        # connection to the pool
        engine_args.setdefault("isolation_level", os.environ.get("CHALK_SQL_ISOLATION_LEVEL", "AUTOCOMMIT"))
        BaseSQLSource.__init__(self, name=name, engine_args=engine_args)

    def local_engine_url(self) -> URL:
        try:
            from sqlalchemy.engine.url import URL
        except ImportError:
            raise missing_dependency_exception("chalkpy[sqlite]")
        database = ":memory:" if self.filename is None else str(self.filename)
        return URL.create(drivername="sqlite", database=database, query={"check_same_thread": "true"})
